# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2016-2019 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
#

import os
import sys
import logging
import subprocess
import shlex
import tempfile
if sys.version_info.major < 3:
    from pipes import quote
else:
    from shlex import quote

from config_system import get_config, get_config_bool, get_config_string, set_config

logger = logging.getLogger(__name__)

def check_output(command, dir=None):
    """Executes the command, while making sure the executable is found in the $PATH,
    and returns the output. If the executable wasn't found, returns an empty string.
    The 'command' needs to be an array of arguments.
    """
    output = ''
    try:
        output = subprocess.check_output(command, cwd=dir).strip()
        output = output.decode(sys.getdefaultencoding())
    except OSError as e:
        logger.error("%s executing '%s'" % (str(e), command[0]))
    except subprocess.CalledProcessError as e:
        logger.warning("Problem executing command: %s" % str(e))

    return output

def check_exit_code(command, dir=None):
    '''
    Executes the command, while making sure the executable is found in the $PATH,
    and returns the exit code. If the executable wasn't found, returns 1.
    The 'command' needs to be an array of arguments.
    '''
    try:
        output = subprocess.check_output(command, cwd=dir, stderr=subprocess.STDOUT).strip()
    except OSError:
        return 1
    except subprocess.CalledProcessError as e:
        logger.info("Determining exit code: %s" % str(e))
        logger.info(e.output)
        return e.returncode

    return 0

def has_compiler_flag(compiler, flag, extra_flags=None):
    '''
    Return 1 when the compiler supports the flag, or 0 when it does not.
    '''
    with tempfile.NamedTemporaryFile() as fp:
        fp.write(b"int main() { return 0; }")
        fp.flush()
        outfile = fp.name + ".o"
        cmd = [compiler, "-c", "-x", "c++"]
        if extra_flags and extra_flags != '':
            cmd.extend(extra_flags.split(' '))
        cmd.extend(['-Werror', flag, fp.name, "-o", outfile])
        if check_exit_code(cmd) == 0:
            return 1

        # Clean up the output file.
        try:
            os.remove(outfile)
        except (OSError) as e:
            # Assume compiler did not support the flag if removal fails.
            return 0

    return 0

def gnu_toolchain_is_cross_compiler(tgtType):
    prefix = get_config_string(tgtType + "_GNU_PREFIX")
    # If the prefix ends with the path separator, it is being used as
    # alternative to adding the compiler to $PATH, and isn't specifying a
    # cross-compilation prefix.
    if prefix and not prefix.endswith(os.sep):
        return True
    return False

def get_cxx_compiler(tgtType):
    """Parse the config options prefixed by `tgtType` and determine the name of
    the compiler for that target type, as well as whether it is a cross
    compiler.
    """
    args = []
    cross_compiler = False

    if get_config_bool(tgtType + "_TOOLCHAIN_GNU"):
        args = shlex.split(get_config_string(tgtType + "_GNU_FLAGS"))
        cxx = get_config_string(tgtType + "_GNU_PREFIX") + get_config_string("GNU_CXX_BINARY")
        cross_compiler = gnu_toolchain_is_cross_compiler(tgtType)
    elif get_config_bool(tgtType + "_TOOLCHAIN_CLANG"):
        # To use the *_CLANG_TRIPLE option to decide whether this is a
        # cross-compiler would require discovering what target triples are
        # runnable on the host. Instead, just ask the GNU toolchain.
        cross_compiler = gnu_toolchain_is_cross_compiler(tgtType)
        cxx = get_config_string(tgtType + "_CLANG_PREFIX") + get_config_string("CLANG_CXX_BINARY")
    elif get_config_bool(tgtType + "_TOOLCHAIN_ARMCLANG"):
        if tgtType != "HOST":
            cross_compiler = True
        cxx = get_config_string(tgtType + "_ARMCLANG_PREFIX") + get_config_string("ARMCLANG_CXX_BINARY")

    return cxx, args, cross_compiler

def stl_rpath_ldflags(tgtType):
    """If the toolchain is generating code that can run on the host, add the
    C++ STL library to the linker rpath so it can be found when the binary is
    executed.
    """

    stl_library = get_config_string(tgtType + "_STL_LIBRARY")
    if not stl_library:
        return ""

    cxx, args, cross_compiler = get_cxx_compiler(tgtType)

    if cross_compiler:
        return ""

    stl_lib = "lib{}.so".format(stl_library)
    stl_path = check_output([cxx] + args + ["-print-file-name=" + stl_lib])
    stl_dir = os.path.dirname(stl_path)
    ld_library_path = os.getenv("LD_LIBRARY_PATH", default="")
    ld_paths = ld_library_path.split(os.pathsep)
    if stl_dir and stl_dir not in ld_paths:
        return "-Wl,--enable-new-dtags " + quote("-Wl,-rpath,{}".format(stl_dir))
    return ""

def get_gcc_sysroot_for_clang(target_type):
    # If Clang is using any components from a GNU toolchain, we can assume `gcc`
    # is available. Ignore CLANG_USE_GNU_BINUTILS, because we could be using a
    # binutils installation which does not contain a C compiler.
    if not (get_config_bool(target_type + "_CLANG_USE_GNU_CRT") or
            get_config_bool(target_type + "_CLANG_USE_GNU_STL") or
            get_config_bool(target_type + "_CLANG_USE_GNU_LIBGCC")):
        return ""

    prefix = get_config_string("%s_GNU_PREFIX" % target_type)
    gcc = prefix + get_config_string("GNU_CC_BINARY")
    try:
        cmd = [gcc, "--print-sysroot"]
        sysroot = subprocess.check_output(cmd).decode().strip()
    except (OSError, UnicodeDecodeError, subprocess.CalledProcessError) as e:
        logger.warning("Could not get sysroot from GCC binary %s: %s", gcc, e)
        return ""

    if not os.path.isdir(sysroot):
        logger.warning("Sysroot '%s' reported by %s does not exist: ignoring", sysroot, gcc)

    return sysroot

def get_cxx_binary(target_type):
    for toolchain in ['GNU', 'CLANG', 'ARMCLANG']:
        if get_config_bool('%s_TOOLCHAIN_%s' % (target_type, toolchain)):
            return get_config_string('%s_CXX_BINARY' % (toolchain))
    logger.error("Couldn't get CXX binary for target type %s", target_type)
    return "g++"

def compiler_config():
    '''
    Adds runtime-dependant entries by analyzing user-configured flags.
    This is mainly required to allow some configuration-specific flags that are
    impossible or hard to express in Blueprint, because of the need to, say, run
    a binary and get the output from it.
    '''
    host_cxx = get_cxx_binary("HOST")
    target_cxx = get_cxx_binary("TARGET")
    target_cxx_flags = None
    # No toolchain prefix indicates a native build
    native_build = get_config_string('TARGET_GNU_PREFIX') == ''
    if native_build:
        target_cxx = host_cxx
    else:
        if get_config_bool('TARGET_TOOLCHAIN_CLANG'):
            target_cxx_flags = '-target ' + get_config_string("TARGET_CLANG_TRIPLE")

            # If the user hasn't set a value, choose a default sysroot.
            sysroot = get_config('TARGET_SYSROOT')
            if not sysroot['is_user_set']:
                set_config('TARGET_SYSROOT', get_gcc_sysroot_for_clang('TARGET'), is_user_set=False)
        elif get_config_bool('TARGET_TOOLCHAIN_GNU'):
            target_cxx_flags = get_config_string('TARGET_GNU_FLAGS')

    # The LLVM libraries have no stable C++ API, so -Wnoexcept-type is not useful.
    # But only add -Wno-noexcept-type if the compiler supports it.
    if has_compiler_flag(target_cxx, '-Werror=noexcept-type', target_cxx_flags):
        set_config('EXTRA_LLVM_TARGET_CXXFLAGS', '-Wno-noexcept-type')
    if not native_build:
        if has_compiler_flag(host_cxx, '-Werror=noexcept-type'):
            set_config('EXTRA_LLVM_HOST_CXXFLAGS', '-Wno-noexcept-type')

    # Extra LDFLAGS are set to support library paths needed by some compilers
    set_config("EXTRA_HOST_LDFLAGS", stl_rpath_ldflags("HOST"))
    set_config("EXTRA_TARGET_LDFLAGS", stl_rpath_ldflags("TARGET"))


def plugin_exec():
    if get_config_bool('ALLOW_HOST_EXPLORE'):
        compiler_config()
